from pypresence import Client
from pypresence.types import ActivityType, StatusDisplayType
import time
import logging

log = logging.getLogger("rich_presence")

client_id = "743296148592263240"
RPC = None


def _ensure_connected():
    global RPC
    if RPC is not None:
        return True
    try:
        rpc = Client(client_id)
        rpc.start()
        RPC = rpc
        log.info("Connected to Discord Rich Presence.")
        return True
    except Exception:
        log.debug("Discord is not available yet.", exc_info=True)
        return False


def register_join_event(syncplay_join_group: callable):
    if _ensure_connected():
        RPC.register_event("activity_join", syncplay_join_group)


def send_presence(
    title: str,
    subtitle: str,
    playback_time: float = None,
    duration: float = None,
    playing: bool = False,
    syncplay_group: str = None,
    media_type: str = None,
):
    if not _ensure_connected():
        return

    small_image = "play-dark3" if playing else None
    start = None
    end = None
    if playback_time is not None and duration is not None and playing:
        start = int(time.time() - playback_time)
        end = int(start + duration)

    payload = {
        "activity_type": (
            ActivityType.LISTENING
            if media_type and media_type == "Audio"
            else ActivityType.WATCHING
        ),
        "status_display_type": StatusDisplayType.DETAILS,
        "state": subtitle if subtitle else "Unknown Media",
        "details": title,
        "instance": False,
        "large_image": "jellyfin2",
        "start": start,
        "end": end,
        "large_text": "Jellyfin",
        "small_image": small_image,
    }

    if syncplay_group:
        payload["party_id"] = str(hash(syncplay_group))
        payload["party_size"] = [1, 100]
        payload["join"] = syncplay_group

    try:
        RPC.set_activity(**payload)
    except Exception:
        log.warning("Discord connection lost.", exc_info=True)
        RPC = None


def clear_presence():
    global RPC
    if RPC is None:
        return
    try:
        RPC.clear_activity()
    except Exception:
        log.warning("Discord connection lost.", exc_info=True)
        RPC = None



#!/usr/bin/env python3

import logging
import sys
import multiprocessing
from threading import Event

from . import conffile
from . import i18n
from .args import get_args
from .conf import settings
from .constants import APP_NAME
from .log_utils import (
    configure_log,
    configure_log_file,
    enable_sanitization,
    root_logger,
)

logging.getLogger("requests").setLevel(logging.CRITICAL)


def main():
    args = get_args()

    conf_file = conffile.get(APP_NAME, "conf.json")
    load_success = settings.load(conf_file)
    i18n.configure()

    # CLI overrides applied after config load so they win.
    if args.enable_gui is not None:
        settings.enable_gui = args.enable_gui
    if args.mpv_loglevel is not None:
        settings.mpv_log_level = args.mpv_loglevel

    if settings.sanitize_output:
        enable_sanitization()

    app_log_level = "debug" if args.debug else settings.mpv_log_level
    configure_log(sys.stdout, app_log_level)
    if settings.write_logs:
        log_file = conffile.get(APP_NAME, "log.txt")
        configure_log_file(log_file, app_log_level)

    log = root_logger

    if sys.platform.startswith("darwin"):
        try:
            # Use 'spawn' to avoid Objective-C fork crashes with GUI frameworks.
            # - Python 3.7: default is 'fork' (unsafe with Obj-C)
            # - Python 3.8+: default is 'spawn' (this is a no-op but explicit)
            # - Python 3.14: 'forkserver' also crashes with Obj-C (issue #473)
            multiprocessing.set_start_method("spawn")
        except RuntimeError:
            # Context already set, ignore
            pass

    user_interface = None
    mirror = None
    use_gui = False
    gui_ready = None
    get_webview = lambda: None
    if settings.enable_gui:
        try:
            from .gui_mgr import user_interface

            use_gui = True
            gui_ready = Event()
            user_interface.gui_ready = gui_ready
        except Exception:
            log.warning(
                "Cannot load GUI. Falling back to command line interface.",
                exc_info=True,
            )

    if settings.display_mirroring:
        try:
            from .display_mirror import mirror

            get_webview = mirror.get_webview
        except ImportError:
            mirror = None
            log.warning("Cannot load display mirror.", exc_info=True)

    if not user_interface:
        from .cli_mgr import user_interface

    from .clients import clientManager
    from .player import playerManager
    from .action_thread import actionThread
    from .event_handler import eventHandler
    from .timeline import timelineManager

    clientManager.callback = eventHandler.handle_event
    timelineManager.start()
    playerManager.timeline_trigger = timelineManager.trigger
    actionThread.start()
    playerManager.action_trigger = actionThread.trigger
    playerManager.get_webview = get_webview
    user_interface.open_player_menu = playerManager.menu.show_menu
    eventHandler.mirror = mirror
    user_interface.start()
    user_interface.login_servers()

    if not load_success:
        log.error("Your configuration file is not valid JSON! It has been ignored!")
        log.info("Tip: Open the JSON file in VS Code to see what is wrong.")

    try:
        if mirror:
            user_interface.stop_callback = mirror.stop
            # If the webview runs before the systray icon, it fails.
            if use_gui:
                gui_ready.wait()
            mirror.run()
        else:
            halt = Event()
            user_interface.stop_callback = halt.set
            try:
                while not halt.wait(timeout=1):
                    pass
            except KeyboardInterrupt:
                print("")
                log.info("Stopping services...")
    finally:
        playerManager.terminate()
        timelineManager.stop()
        actionThread.stop()
        clientManager.stop()
        user_interface.stop()


if __name__ == "__main__":
    main()



"""
Power-user language preference rules.

A rule list is evaluated in order; the first rule that matches sets the
audio/subtitle tracks. A rule "matches" only when all of its constraints can
be satisfied — partial matches fall through to the next rule, and if no rule
matches, the existing Jellyfin server defaults apply.

Constraint fields (all reject the rule on no-match):
  type     - "movie" or "series" (item-type filter)
  alang    - mpv-style comma list of audio language priorities
  slang    - same for subtitles
  amatch   - regex over audio track titles
  smatch   - regex over subtitle track titles
  subtype  - "signs" or "full" (uses the bulk_subtitle weight helpers)

Bias fields (narrow the candidate set without rejecting the rule):
  aprefer  - regex over audio track titles, applied after alang
  sprefer  - same for subtitles
"""

from __future__ import annotations

import logging
import re
from dataclasses import asdict, dataclass, fields
from typing import List, Optional, Tuple

log = logging.getLogger("language_config")


@dataclass
class LanguageRule:
    type: Optional[str] = None
    alang: Optional[str] = None
    slang: Optional[str] = None
    aprefer: Optional[str] = None
    sprefer: Optional[str] = None
    amatch: Optional[str] = None
    smatch: Optional[str] = None
    subtype: Optional[str] = None

    def _to_dict(self) -> dict:
        return {k: v for k, v in asdict(self).items() if v is not None}


def parse_language_config(value) -> Optional[List[LanguageRule]]:
    if value is None or value == "null":
        return None
    if not isinstance(value, list):
        log.error("language_config must be a list of rule objects, got %r", type(value))
        return None

    known = {f.name for f in fields(LanguageRule)}
    rules = []
    for i, item in enumerate(value):
        if not isinstance(item, dict):
            log.warning("language_config[%d] is not an object, skipping", i)
            continue
        unknown = set(item.keys()) - known
        if unknown:
            log.warning(
                "language_config[%d] has unknown fields %s (ignored)",
                i,
                sorted(unknown),
            )
        filtered = {k: v for k, v in item.items() if k in known}
        try:
            rules.append(LanguageRule(**filtered))
        except Exception:
            log.warning(
                "language_config[%d] failed to parse, skipping", i, exc_info=True
            )
    return rules


def _track_title(stream: dict) -> str:
    return (stream.get("DisplayTitle") or stream.get("Title") or "").lower()


def _filter_by_lang(streams: list, lang_priority: str) -> Optional[list]:
    """Return the first non-empty bucket of streams matching a priority entry."""
    for lang in lang_priority.split(","):
        lang = lang.strip().lower()
        if not lang:
            continue
        bucket = [s for s in streams if (s.get("Language") or "").lower() == lang]
        if bucket:
            return bucket
    return None


def _bias_by_regex(streams: list, pattern: str) -> list:
    pat = re.compile(pattern, re.IGNORECASE)
    preferred = [s for s in streams if pat.search(_track_title(s))]
    return preferred or streams


def _filter_subtype(streams: list, subtype: str) -> list:
    """signs = sign_weight > 0 OR forced; full = neither sign-y nor forced."""
    # Deferred to break a conf -> language_config -> bulk_subtitle -> utils -> conf cycle.
    from .bulk_subtitle import sign_weight

    if subtype == "signs":
        return [
            s for s in streams if s.get("IsForced") or sign_weight(_track_title(s)) > 0
        ]
    if subtype == "full":
        return [
            s
            for s in streams
            if not s.get("IsForced") and sign_weight(_track_title(s)) == 0
        ]
    log.warning("language_config: unknown subtype %r (ignored)", subtype)
    return streams


def _pick_best_sub(streams: list, subtype: Optional[str]):
    """Among candidates, pick the one most appropriate to the requested intent.

    For subtype="signs": lowest sign_weight wins (most clearly a signs track).
    Otherwise: lowest dialogue_weight wins (most clearly a full-dialogue track).
    Ties broken by source order.
    """
    if not streams:
        return None
    from .bulk_subtitle import dialogue_weight, sign_weight

    score = sign_weight if subtype == "signs" else dialogue_weight
    return min(streams, key=lambda s: score(_track_title(s)))


_TYPE_MAP = {"movie": {"Movie"}, "series": {"Episode"}}


def _matches_type(item: dict, rule_type: Optional[str]) -> bool:
    if not rule_type:
        return True
    expected = _TYPE_MAP.get(rule_type.lower())
    if expected is None:
        log.warning("language_config: unknown type %r (rule skipped)", rule_type)
        return False
    return item.get("Type") in expected


def _try_match(
    rule: LanguageRule, source: dict, item: dict
) -> Optional[Tuple[Optional[int], Optional[int]]]:
    """Return (aid, sid) if the rule matches, else None. Either aid or sid may be None."""
    if not _matches_type(item, rule.type):
        return None

    streams = source.get("MediaStreams") or []
    audio = [s for s in streams if s.get("Type") == "Audio"]
    subs = [s for s in streams if s.get("Type") == "Subtitle"]

    # Audio constraints
    if rule.amatch:
        try:
            pat = re.compile(rule.amatch, re.IGNORECASE)
        except re.error:
            log.warning("language_config: invalid amatch regex %r", rule.amatch)
            return None
        audio = [s for s in audio if pat.search(_track_title(s))]
        if not audio:
            return None

    chosen_audio = None
    if rule.alang:
        bucket = _filter_by_lang(audio, rule.alang)
        if bucket is None:
            return None
        if rule.aprefer:
            bucket = _bias_by_regex(bucket, rule.aprefer)
        chosen_audio = bucket[0]
    elif rule.aprefer:
        bucket = _bias_by_regex(audio, rule.aprefer)
        chosen_audio = bucket[0] if bucket else None

    # Subtitle constraints
    if rule.subtype:
        subs = _filter_subtype(subs, rule.subtype)
        if not subs:
            return None
    if rule.smatch:
        try:
            pat = re.compile(rule.smatch, re.IGNORECASE)
        except re.error:
            log.warning("language_config: invalid smatch regex %r", rule.smatch)
            return None
        subs = [s for s in subs if pat.search(_track_title(s))]
        if not subs:
            return None

    chosen_sub = None
    if rule.slang:
        bucket = _filter_by_lang(subs, rule.slang)
        if bucket is None:
            return None
        if rule.sprefer:
            bucket = _bias_by_regex(bucket, rule.sprefer)
        chosen_sub = _pick_best_sub(bucket, rule.subtype)
    elif rule.sprefer:
        bucket = _bias_by_regex(subs, rule.sprefer)
        chosen_sub = _pick_best_sub(bucket, rule.subtype)

    aid = chosen_audio["Index"] if chosen_audio else None
    sid = chosen_sub["Index"] if chosen_sub else None
    return aid, sid


def apply(
    rules: Optional[List[LanguageRule]], source: dict, item: dict
) -> Tuple[Optional[int], Optional[int]]:
    """Walk the rule list; return (aid, sid) from the first matching rule, or (None, None)."""
    if not rules or not source:
        return None, None
    for i, rule in enumerate(rules):
        result = _try_match(rule, source, item)
        if result is not None:
            log.info("matched rule %d -> aid=%s, sid=%s", i + 1, *result)
            return result
    log.info("no rule matched")
    return None, None



import threading
import os
import logging
import math

from .conf import settings
from . import conffile
from .constants import APP_NAME

try:
    from . import bifdecode

    BIFDECODE_AVAILABLE = bifdecode.PIL_AVAILABLE
except ImportError:
    BIFDECODE_AVAILABLE = False
    bifdecode = None

log = logging.getLogger("trickplay")
img_file = conffile.get(APP_NAME, "raw_images.bin")


class TrickPlay(threading.Thread):
    def __init__(self, player):
        self.trigger = threading.Event()
        self.halt = False
        self.player = player

        threading.Thread.__init__(self)

    def stop(self):
        self.halt = True
        self.trigger.set()
        if os.path.isfile(img_file):
            os.remove(img_file)
        self.join()

    def fetch_thumbnails(self):
        self.trigger.set()

    def clear(self):
        self.player.script_message("shim-trickplay-clear")
        if os.path.isfile(img_file):
            os.remove(img_file)

    def run(self):
        if not BIFDECODE_AVAILABLE:
            log.warning(
                "Trickplay thumbnails disabled: Pillow (PIL) not available. Install with: pip install pillow"
            )
            return

        while not self.halt:
            self.trigger.wait()
            self.trigger.clear()

            if self.halt:
                break

            try:
                log.info("Collecting trickplay images...")

                if not self.player.has_video():
                    continue

                video = self.player.get_video()
                try:
                    data = video.get_bif(settings.thumbnail_preferred_size)
                    if not self.player.has_video() or video != self.player.get_video():
                        # Video changed while we were getting the bif file
                        continue

                    if data:
                        with open(img_file, "wb") as fh:
                            img_count = math.ceil(
                                data["ThumbnailCount"]
                                / data["TileWidth"]
                                / data["TileHeight"]
                            )
                            bifdecode.decompress_tiles(
                                data["Width"],
                                data["Height"],
                                data["TileWidth"],
                                data["TileHeight"],
                                data["ThumbnailCount"],
                                video.get_hls_tile_images(data["Width"], img_count),
                                fh,
                            )

                        bif_meta = {
                            "count": data["ThumbnailCount"],
                            "multiplier": data["Interval"],
                            "width": data["Width"],
                            "height": data["Height"],
                        }

                        if (
                            not self.player.has_video()
                            or video != self.player.get_video()
                        ):
                            # Video changed while we were decompressing the bif file
                            continue

                        self.player.script_message(
                            "shim-trickplay-bif",
                            str(bif_meta["count"]),
                            str(bif_meta["multiplier"]),
                            str(bif_meta["width"]),
                            str(bif_meta["height"]),
                            img_file,
                        )
                        log.info(
                            f"Collected {bif_meta['count']} trickplay preview images"
                        )
                        continue
                    else:
                        log.warning("No trickplay data available")
                except:
                    log.error(
                        "Could not get trickplay data.",
                        exc_info=True,
                    )

                chapter_data = video.get_chapters()

                if chapter_data is None or len(chapter_data) == 0:
                    log.info("No chapters available")
                    continue

                with open(img_file, "wb") as fh:
                    bif_meta = bifdecode.decompress_bif(
                        video.get_chapter_images(settings.thumbnail_preferred_size), fh
                    )

                if not self.player.has_video() or video != self.player.get_video():
                    # Video changed while we were getting the thumbnails
                    break

                self.player.script_message(
                    "shim-trickplay-chapters",
                    str(bif_meta["width"]),
                    str(bif_meta["height"]),
                    str(img_file),
                    ",".join(str(x["start"]) for x in chapter_data),
                )
                log.info(f"Collected {len(chapter_data)} chapter preview images")

            except:
                log.error("Could not get trickplay images", exc_info=True)



#!/usr/bin/env python3
from io import BytesIO

try:
    from PIL import Image

    PIL_AVAILABLE = True
except ImportError:
    PIL_AVAILABLE = False
    Image = None


def decompress_tiles(width, height, tile_width, tile_height, count, tiles, fh):
    if not PIL_AVAILABLE:
        raise ImportError(
            "Pillow (PIL) is required for trickplay thumbnails. Install with: pip install pillow"
        )

    image_count = 0

    for image in tiles:
        image = Image.open(BytesIO(image)).convert("RGBA")

        if height * tile_height != image.height or width * tile_width != image.width:
            raise ValueError("Tile size mismatch.")

        r, g, b, a = image.split()
        image_data = Image.merge("RGBA", (b, g, r, a)).tobytes()

        for y in range(tile_height):
            for x in range(tile_width):
                if image_count >= count:
                    return
                image_count += 1

                for y_local in range(height):
                    position = (
                        y * height * width * tile_width * 4  # seek to correct row
                        + x * width * 4  # seek to correct column
                        + y_local * width * tile_width * 4  # seek to correct subrow
                    )
                    fh.write(image_data[position : position + width * 4])


def decompress_bif(images, fh):
    if not PIL_AVAILABLE:
        raise ImportError(
            "Pillow (PIL) is required for trickplay thumbnails. Install with: pip install pillow"
        )

    height = None
    width = None
    image_count = 0

    for image in images:
        image_count += 1
        image = Image.open(BytesIO(image)).convert("RGBA")
        if height is None:
            height = image.height
            width = image.width
        else:
            if height != image.height or width != image.width:
                raise ValueError("BIF image sizes mismatch.")

        r, g, b, a = image.split()
        image = Image.merge("RGBA", (b, g, r, a))

        fh.write(image.tobytes())

    return {"count": image_count, "height": height, "width": width}



"""Tkinter-based fullscreen "ready to cast" / item-preview window.

Replaces the previous pywebview + Jinja2 HTML implementation. Same public
surface as the old DisplayMirror (run/stop/display_content/get_webview), so
the rest of the app needs no changes.

Design:
- A single fullscreen Tk root with a Canvas.
- Backdrop image fetched via requests, scaled to cover, then composed with
  a vertical dark gradient via Pillow for text legibility.
- Title / misc info / overview drawn as Canvas text items, positioned by
  bbox(...) walking up from the bottom-left.
- All Tk mutations happen on the main thread via root.after; commands from
  other threads (websocket events, player playback start/stop) are
  marshalled through a queue.Queue.
"""

import datetime
import logging
import math
import queue as queue_mod
import random
import threading
import tkinter as tk
from io import BytesIO
from typing import TYPE_CHECKING, Optional

import requests
from PIL import Image, ImageTk

from .clients import clientManager
from .i18n import _

log = logging.getLogger("display_mirror")

if TYPE_CHECKING:
    from jellyfin_apiclient_python import client as client_type


# ---- URL builders (ported from the old display_mirror.helpers) ---------

def _server_url(server: str, path: str) -> str:
    return f"{server.rstrip('/')}/{path.lstrip('/')}"


def _backdrop_url(item: dict, server: str) -> Optional[str]:
    if item.get("BackdropImageTags"):
        return _server_url(
            server,
            f"Items/{item['Id']}/Images/Backdrop/0?tag={item['BackdropImageTags'][0]}",
        )
    if item.get("ParentBackdropItemId"):
        return _server_url(
            server,
            f"Items/{item['ParentBackdropItemId']}/Images/Backdrop/0"
            f"?tag={item['ParentBackdropImageTags'][0]}",
        )
    return None


def _logo_url(item: dict, server: str) -> Optional[str]:
    if item.get("ImageTags", {}).get("Logo"):
        return _server_url(
            server, f"Items/{item['Id']}/Images/Logo/0?tag={item['ImageTags']['Logo']}"
        )
    if item.get("ParentLogoItemId") and item.get("ParentLogoImageTag"):
        return _server_url(
            server,
            f"Items/{item['ParentLogoItemId']}/Images/Logo/0"
            f"?tag={item['ParentLogoImageTag']}",
        )
    return None


def _display_name(item: dict) -> str:
    name = item.get("EpisodeTitle") or item.get("Name", "")
    if item.get("Type") == "TvChannel":
        if item.get("Number"):
            return f"{item['Number']} {name}"
        return name
    if (
        item.get("Type") == "Episode"
        and item.get("IndexNumber") is not None
        and item.get("ParentIndexNumber") is not None
    ):
        number = f"S{item['ParentIndexNumber']} E{item['IndexNumber']}"
        if item.get("IndexNumberEnd"):
            number += f"-{item['IndexNumberEnd']}"
        name = f"{number} - {name}"
    return name


def _parse_jf_datetime(s: str) -> datetime.datetime:
    return datetime.datetime.strptime(s.partition(".")[0], "%Y-%m-%dT%H:%M:%S")


def _misc_info(item: dict) -> str:
    parts = []
    typ = item.get("Type")

    if typ == "Episode" and item.get("PremiereDate"):
        parts.append(_parse_jf_datetime(item["PremiereDate"]).strftime("%x"))
    if item.get("StartDate"):
        parts.append(_parse_jf_datetime(item["StartDate"]).strftime("%x"))
    if typ == "Series" and item.get("ProductionYear"):
        if item.get("Status") == "Continuing":
            parts.append(f"{item['ProductionYear']}-Present")
        else:
            text = str(item["ProductionYear"])
            if item.get("EndDate"):
                end_year = _parse_jf_datetime(item["EndDate"]).year
                if end_year != item["ProductionYear"]:
                    text += f"-{end_year}"
            parts.append(text)
    if typ not in ("Series", "Episode"):
        if item.get("ProductionYear"):
            parts.append(str(item["ProductionYear"]))
        elif item.get("PremiereDate"):
            parts.append(str(_parse_jf_datetime(item["PremiereDate"]).year))
    if item.get("RunTimeTicks") and typ not in ("Series", "Audio"):
        minutes = math.ceil(item["RunTimeTicks"] / 600000000)
        parts.append(f"{minutes}min")
    if item.get("OfficialRating") and typ not in ("Season", "Episode"):
        parts.append(item["OfficialRating"])
    if item.get("Video3DFormat"):
        parts.append("3D")
    return "    ".join(parts)


def _rating(item: dict) -> str:
    rating = item.get("CommunityRating")
    if rating:
        return f"★ {rating:.1f}"
    return ""


def _random_backdrop_url() -> Optional[str]:
    if not clientManager.clients:
        return None
    try:
        client = random.choice(list(clientManager.clients.values()))
        params = {
            "SortBy": "Random",
            "Limit": 1,
            "IncludeItemTypes": "Movie,Series",
            "ImageTypes": "Backdrop",
            "Recursive": True,
            "MaxOfficialRating": "PG-13",
        }
        items = client.jellyfin.user_items(params=params).get("Items") or []
        if not items:
            return None
        return _backdrop_url(items[0], client.config.data["auth.server"])
    except Exception:
        log.warning("Could not fetch random backdrop.", exc_info=True)
        return None


# ---- Image processing ---------------------------------------------------

def _fetch_image(url: Optional[str], timeout: int = 10) -> Optional["Image.Image"]:
    if not url:
        return None
    try:
        r = requests.get(url, timeout=timeout)
        r.raise_for_status()
        return Image.open(BytesIO(r.content)).convert("RGBA")
    except Exception:
        log.warning("Failed to fetch image %s", url, exc_info=True)
        return None


def _scale_to_cover(image: "Image.Image", w: int, h: int) -> "Image.Image":
    """Scale `image` to fully cover (w, h), center-cropping the overflow."""
    iw, ih = image.size
    scale = max(w / iw, h / ih)
    new_w, new_h = max(1, int(iw * scale)), max(1, int(ih * scale))
    image = image.resize((new_w, new_h), Image.LANCZOS)
    left = (new_w - w) // 2
    top = (new_h - h) // 2
    return image.crop((left, top, left + w, top + h))


def _apply_dark_gradient(
    image: "Image.Image", height_fraction: float = 0.55, max_alpha: int = 200
) -> "Image.Image":
    """Composite a vertical transparent->dark gradient over the image's bottom."""
    w, h = image.size
    grad_h = max(1, int(h * height_fraction))
    overlay = Image.new("RGBA", (w, h), (0, 0, 0, 0))
    # Build the gradient as a single column then resize horizontally — much
    # faster than per-row paste for large images.
    column = Image.new("RGBA", (1, grad_h))
    for y in range(grad_h):
        alpha = int(max_alpha * (y / max(1, grad_h - 1)) ** 1.5)
        column.putpixel((0, y), (0, 0, 0, alpha))
    column = column.resize((w, grad_h), Image.NEAREST)
    overlay.paste(column, (0, h - grad_h))
    return Image.alpha_composite(image, overlay)


# ---- The window ---------------------------------------------------------

class DisplayMirror:
    def __init__(self):
        self.queue: queue_mod.Queue = queue_mod.Queue()
        self.root: Optional[tk.Tk] = None
        self.canvas: Optional[tk.Canvas] = None
        self.screen_w = 0
        self.screen_h = 0
        # Tk PhotoImage is GC'd if not held; keep the most recent reference.
        self._bg_photo: Optional[ImageTk.PhotoImage] = None

    # --- public API matching the previous DisplayMirror -----------------

    def get_webview(self):
        return self  # exposes hide()/show() for player.py

    def hide(self):
        self.queue.put(("hide", None))

    def show(self):
        self.queue.put(("show", None))

    def stop(self):
        self.queue.put(("die", None))

    def display_content(self, client: "client_type", arguments: dict):
        # Called from the websocket thread. Item fetch is synchronous here
        # (matches the prior behaviour); image fetch is deferred onto a
        # worker thread spawned by the Tk loop.
        try:
            item = client.jellyfin.get_item(arguments["Arguments"]["ItemId"])
        except Exception:
            log.warning("Could not fetch item for display.", exc_info=True)
            return
        server = client.config.data["auth.server"]
        self.queue.put(("display", self._build_item_data(item, server)))

    def run(self):
        self.root = tk.Tk()
        self.root.title("Jellyfin MPV Shim Mirror")
        self.root.configure(bg="black")
        self.root.attributes("-fullscreen", True)
        self.root.config(cursor="none")

        self.screen_w = self.root.winfo_screenwidth()
        self.screen_h = self.root.winfo_screenheight()

        self.canvas = tk.Canvas(
            self.root,
            width=self.screen_w,
            height=self.screen_h,
            bg="black",
            highlightthickness=0,
        )
        self.canvas.pack(fill="both", expand=True)

        # Show idle state on startup. Network call goes on a worker so we
        # don't block mainloop's startup.
        threading.Thread(target=self._show_idle_async, daemon=True).start()

        self.root.after(50, self._poll_queue)
        self.root.mainloop()

    # --- internals ------------------------------------------------------

    @staticmethod
    def _build_item_data(item: dict, server: str) -> dict:
        return {
            "title": _display_name(item),
            "overview": item.get("Overview", "") or "",
            "misc": _misc_info(item),
            "rating": _rating(item),
            "backdrop_url": _backdrop_url(item, server),
            "logo_url": _logo_url(item, server),
        }

    def _poll_queue(self):
        try:
            while True:
                cmd, payload = self.queue.get_nowait()
                if cmd == "die":
                    self.root.destroy()
                    return
                if cmd == "hide":
                    self.root.withdraw()
                elif cmd == "show":
                    self.root.deiconify()
                    self.root.attributes("-fullscreen", True)
                elif cmd == "display":
                    threading.Thread(
                        target=self._render_item, args=(payload,), daemon=True
                    ).start()
                elif cmd == "idle":
                    threading.Thread(target=self._show_idle_async, daemon=True).start()
        except queue_mod.Empty:
            pass
        self.root.after(50, self._poll_queue)

    def _show_idle_async(self):
        url = _random_backdrop_url()
        backdrop = _fetch_image(url) if url else None
        data = {
            "title": _("Ready to cast"),
            "overview": _("Select your media in Jellyfin and play it here."),
            "misc": "",
            "rating": "",
        }
        self.root.after(0, lambda: self._render(data, backdrop))

    def _render_item(self, data: dict):
        backdrop = _fetch_image(data.get("backdrop_url"))
        self.root.after(0, lambda: self._render(data, backdrop))

    def _canvas_dimensions(self) -> tuple:
        """Return the actual rendered canvas size, not the (possibly multi-
        monitor) screen size. winfo_screenwidth/height returns the combined
        virtual desktop on multi-monitor and Wayland setups, which causes text
        wrap to misbehave."""
        self.canvas.update_idletasks()
        w = self.canvas.winfo_width()
        h = self.canvas.winfo_height()
        if w <= 1 or h <= 1:
            # Canvas not realized yet — fall back to the root window, then to
            # the (potentially wrong) screen dims as a last resort.
            w = self.root.winfo_width() or self.screen_w
            h = self.root.winfo_height() or self.screen_h
        return w, h

    def _render(self, data: dict, backdrop: Optional["Image.Image"]):
        if not self.canvas:
            return
        self.canvas.delete("all")
        cw, ch = self._canvas_dimensions()

        if backdrop is not None:
            try:
                bg = _scale_to_cover(backdrop, cw, ch)
                bg = _apply_dark_gradient(bg)
                self._bg_photo = ImageTk.PhotoImage(bg)
                self.canvas.create_image(0, 0, image=self._bg_photo, anchor="nw")
            except Exception:
                log.warning("Failed to render backdrop.", exc_info=True)
                self._bg_photo = None

        self._render_text(data, cw, ch)

    def _render_text(self, data: dict, cw: int, ch: int):
        margin = max(40, cw // 30)
        wrap = cw - 2 * margin

        # Font sizes scale with the rendered canvas height; clamped to
        # sensible bounds for very small/large windows.
        info_size = max(14, min(28, ch // 50))
        body_size = max(18, min(36, ch // 30))

        # Title font scales down for long titles so multi-line episode names
        # don't dominate the screen. Hard-cap absurdly long ones.
        title = (data.get("title") or "")[:200]
        base_title_size = max(36, min(96, ch // 14))
        if len(title) > 60:
            title_size = int(base_title_size * 0.6)
        elif len(title) > 40:
            title_size = int(base_title_size * 0.75)
        else:
            title_size = base_title_size

        # Build bottom-up: place each line at the previous line's top edge.
        anchor_y = ch - margin

        def stack_text(text: str, font, fill: str, *, gap: int = 18) -> int:
            nonlocal anchor_y
            if not text:
                return anchor_y
            text_id = self.canvas.create_text(
                margin,
                anchor_y,
                text=text,
                font=font,
                fill=fill,
                anchor="sw",
                width=wrap,
            )
            bbox = self.canvas.bbox(text_id)
            if bbox:
                anchor_y = bbox[1] - gap
            return anchor_y

        stack_text(data.get("overview", ""), ("Helvetica", body_size), "#dddddd")
        misc_line = "    ".join(
            s for s in (data.get("misc"), data.get("rating")) if s
        )
        stack_text(misc_line, ("Helvetica", info_size), "#bbbbbb")
        stack_text(title, ("Helvetica", title_size, "bold"), "#ffffff", gap=8)


mirror = DisplayMirror()



from jellyfin_apiclient_python import JellyfinClient
from jellyfin_apiclient_python.connection_manager import CONNECTION_STATE
from .conf import settings
from . import conffile
from getpass import getpass
from .constants import CAPABILITIES, CLIENT_VERSION, USER_APP_NAME, USER_AGENT, APP_NAME
from .i18n import _

import os.path
import json
import uuid
import time
import logging
import re
import threading

import socket
import ipaddress
from urllib.parse import urlparse


# Get all local IPv4 addresses for host machine
def get_local_ips():
    local_ips = []
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(("8.8.8.8", 80))
            local_ips.append(ipaddress.ip_address(s.getsockname()[0]))
    except OSError:
        pass

    try:
        for info in socket.getaddrinfo(socket.gethostname(), None, socket.AF_INET):
            ip = ipaddress.ip_address(info[4][0])
            if ip not in local_ips and ip.is_private:
                local_ips.append(ip)
    except socket.gaierror:
        pass

    return local_ips


# Extract hostname/IP from a URL.
def extract_host(url):
    parsed = urlparse(url)
    return parsed.hostname  # Handles port stripping automatically


# Check if string is valid IPv4 address
def is_ipv4(host):
    try:
        ipaddress.IPv4Address(host)
        return True
    except ipaddress.AddressValueError:
        return False


# Check if host IP is on the same subnet as any local IP
def is_local_subnet(host, local_ips, prefix_length=24):
    try:
        target_ip = ipaddress.IPv4Address(host)
    except ipaddress.AddressValueError:
        return False

    if not target_ip.is_private:
        return False

    for local_ip in local_ips:
        net = ipaddress.ip_network(f"{local_ip}/{prefix_length}", strict=False)
        if target_ip in net:
            return True
    return False


log = logging.getLogger("clients")
path_regex = re.compile(r"^(https?://)?(?:(\[[^/]+\])|([^/:]+))(:[0-9]+)?(/.*)?$")

from typing import Optional


def expo(max_value: Optional[int] = None):
    n = 0
    while True:
        a = 2**n
        if max_value is None or a < max_value:
            yield a
            n += 1
        else:
            yield max_value


class PeriodicHealthCheck(threading.Thread):
    def __init__(self, callback):
        self.halt = False
        self.trigger = threading.Event()
        self.callback = callback

        threading.Thread.__init__(self, daemon=True)

    def stop(self):
        self.halt = True
        self.trigger.set()
        self.join()

    def run(self):
        while not self.halt:
            if not self.trigger.wait(settings.health_check_interval):
                self.callback()


class ClientManager(object):
    def __init__(self):
        self.callback = lambda client, event_name, data: None
        self.credentials = []
        self.clients = {}
        self.usernames = {}
        self.is_stopping = False

        self.health_check = None
        if settings.health_check_interval is not None:
            self.health_check = PeriodicHealthCheck(self.check_all_clients)
            self.health_check.start()

    @staticmethod
    def _get_cli_credential_args():
        from .args import get_args
        a = get_args()
        if a.server and a.username:
            return a.server, a.username, a.password
        return None

    def _find_existing_credential(self, server: str, username: str):
        """Find an existing credential matching the given server and username."""
        if server.endswith("/"):
            server = server[:-1]
        for cred in self.credentials:
            cred_address = cred.get("address", "").rstrip("/")
            cred_username = cred.get("username", "")
            if cred_address == server and cred_username == username:
                return cred
        return None

    def _update_account(self, server: str, username: str, password: str):
        """Update an existing account by re-authenticating with new credentials."""
        existing = self._find_existing_credential(server, username)
        if existing is None:
            return False

        # Disconnect the old client
        self._disconnect_client(uuid=existing["uuid"])
        # Remove the old credential
        self.credentials = [
            c for c in self.credentials if c["uuid"] != existing["uuid"]
        ]
        self.save_credentials()

        # Re-login with updated credentials
        return self.login(server, username, password)

    def cli_connect(self):
        from .args import get_args
        cli_commands = set(get_args().command or [])

        is_logged_in = self.try_connect()
        add_another = "add" in cli_commands
        clear_accounts = "clear" in cli_commands

        cli_creds = self._get_cli_credential_args()

        if clear_accounts:
            log.info(_("Clearing all existing accounts."))
            self.remove_all_clients()
            is_logged_in = False

        if cli_creds:
            server, username, password = cli_creds
            if not clear_accounts:
                existing = self._find_existing_credential(server, username)
                if existing is not None:
                    log.info(_("Account already exists. Updating credentials."))
                    if self._update_account(server, username, password):
                        log.info(_("Successfully updated server credentials."))
                        is_logged_in = True
                    else:
                        log.warning(_("Updating server credentials failed."))
                    cli_creds = None

            if cli_creds and (not is_logged_in or add_another or clear_accounts):
                is_logged_in = self.login(server, username, password)
                if is_logged_in:
                    log.info(_("Successfully added server."))
                else:
                    log.warning(_("Adding server failed."))

        while not is_logged_in or add_another:
            server = input(_("Server URL: "))
            username = input(_("Username: "))
            try:
                password = getpass(_("Password: "))
            except (EOFError, OSError):
                password = ""

            is_logged_in = self.login(server, username, password)

            if is_logged_in:
                log.info(_("Successfully added server."))
                add_another = input(_("Add another server?") + " [y/N] ")
                add_another = add_another in ("y", "Y", "yes", "Yes")
            else:
                log.warning(_("Adding server failed."))

    @staticmethod
    def client_factory():
        client = JellyfinClient(allow_multiple_clients=True)
        client.config.data["app.default"] = True
        client.config.app(
            USER_APP_NAME, CLIENT_VERSION, settings.player_name, settings.client_uuid
        )
        client.config.data["http.user_agent"] = USER_AGENT
        client.config.data["auth.ssl"] = not settings.ignore_ssl_cert

        if settings.tls_client_cert:
            client.config.data["auth.tls_client_cert"] = settings.tls_client_cert
            client.config.data["auth.tls_client_key"] = settings.tls_client_key
            client.config.data["auth.tls_server_ca"] = settings.tls_server_ca
            client.auth.create_session_with_client_auth()

        return client

    def _connect_all(self):
        is_logged_in = False

        local_ips = get_local_ips()

        def connection_priority(server):
            host = extract_host(server["address"])

            # Highest priority: same subnet as us
            if is_ipv4(host) and is_local_subnet(host, local_ips):
                return 0

            # Second priority: other private IPs (maybe reachable via VPN, etc.)
            if is_ipv4(host):
                try:
                    if ipaddress.IPv4Address(host).is_private:
                        return 1
                except ipaddress.AddressValueError:
                    pass

            # Lowest priority: hostnames / external addresses
            return 2

        # Sort creds list by local-first priority
        sorted_credentials = sorted(self.credentials, key=connection_priority)

        # Array to stash server Ids, to avoid double-connecting to servers
        # and avoid clobbering the preferred connection
        connected_servers = []

        for server in sorted_credentials:
            # Test if we've connected to this server already
            if server["Id"] in connected_servers:
                # If so, skip connecting
                continue
            if self.connect_client(server):
                is_logged_in = True

                # If valid connection, add Id of server to array
                connected_servers.append(server["Id"])
        return is_logged_in

    def try_connect(self):
        credentials_location = conffile.get(APP_NAME, "cred.json")
        if os.path.exists(credentials_location):
            with open(credentials_location) as cf:
                self.credentials = json.load(cf)

        if "Servers" in self.credentials:
            credentials_old = self.credentials
            self.credentials = []
            for server in credentials_old["Servers"]:
                server["uuid"] = str(uuid.uuid4())
                server["username"] = ""
                self.credentials.append(server)

        is_logged_in = self._connect_all()
        if settings.connect_retry_mins and not is_logged_in:
            log.warning(
                "Connection failed. Will retry for {0} minutes.".format(
                    settings.connect_retry_mins
                )
            )
            for attempt in range(settings.connect_retry_mins * 2):
                time.sleep(30)
                is_logged_in = self._connect_all()
                if is_logged_in:
                    break

        return is_logged_in

    def save_credentials(self):
        credentials_location = conffile.get(APP_NAME, "cred.json")
        with open(credentials_location, "w") as cf:
            json.dump(self.credentials, cf)

    def login(
        self, server: str, username: str, password: str, force_unique: bool = False
    ):
        if server.endswith("/"):
            server = server[:-1]

        protocol, ipv6_host, ipv4_host, port, path = path_regex.match(server).groups()

        if not protocol:
            log.warning("Adding http:// because it was not provided.")
            protocol = "http://"

        if protocol == "http://" and not port:
            log.warning("Adding port 8096 for insecure local http connection.")
            log.warning(
                "If you want to connect to standard http port 80, use :80 in the url."
            )
            port = ":8096"

        server = "".join(filter(bool, (protocol, ipv6_host, ipv4_host, port, path)))

        client = self.client_factory()
        client.auth.connect_to_address(server)
        result = client.auth.login(server, username, password)
        if "AccessToken" in result:
            credentials = client.auth.credentials.get_credentials()
            server = credentials["Servers"][0]
            if force_unique:
                server["uuid"] = server["Id"]
            else:
                server["uuid"] = str(uuid.uuid4())
            server["username"] = username
            if force_unique and server["Id"] in self.clients:
                return True
            self.connect_client(server)
            self.credentials.append(server)
            self.save_credentials()
            return True
        return False

    def validate_client(self, client: "JellyfinClient", dry_run=False):
        # Use the apiclient's lower-level _http to bound retries and timeout
        # for this specific call. The default 30s × 5 retries can wedge the
        # health-check thread for ~2.5 minutes if the server is unresponsive.
        # On exception, fall through to the "not in client list" branch below
        # to force a reconnect (a timeout is a broken connection).
        try:
            client_list = client.jellyfin._http(
                "GET", "Sessions", {"params": None, "timeout": 10, "retry": 1}
            )
        except Exception:
            log.warning("Health check session query failed; treating as disconnected.", exc_info=True)
            client_list = []

        if client_list is None:
            log.warning(
                "Client check failed, proceeding anyways. (Client list is unset.)"
            )
            return True

        for f_client in client_list:
            if f_client.get("DeviceId") == settings.client_uuid:
                break
        else:
            if not dry_run:
                log.warning(
                    "Client is not actually connected. (It does not show in the client list.)"
                )
                # WebSocketDisconnect doesn't always happen here.
                client.callback = lambda *_: None
                client.callback_ws = lambda *_: None
                client.stop()
                client.callback("WebSocketDisconnect", None)
            return False

        return True

    def setup_client(self, client: "JellyfinClient", server):
        def event(event_name, data):
            if event_name == "WebSocketDisconnect":
                timeout_gen = expo(100)
                if server["uuid"] in self.clients:
                    while not self.is_stopping:
                        timeout = next(timeout_gen)
                        log.info(
                            "No connection to server. Next try in {0} second(s)".format(
                                timeout
                            )
                        )
                        self._disconnect_client(server=server)
                        time.sleep(timeout)
                        if self.connect_client(server, False):
                            break
            elif event_name == "WebSocketConnect":
                log.info("WebSocket connected, posting capabilities")
                try:
                    client.jellyfin.post_capabilities(CAPABILITIES)
                except Exception:
                    log.warning(
                        "Failed to post capabilities on reconnect", exc_info=True
                    )
                self.callback(client, event_name, data)
            else:
                self.callback(client, event_name, data)

        client.callback = event
        client.callback_ws = event
        client.start(websocket=True)

        # Check connection
        if self.validate_client(client, True):
            return True

        # Wait and check connection again before destroying/re-creating client
        log.info("Not connected yet, waiting 3 seconds...")
        time.sleep(3)
        is_connected = self.validate_client(client)

        if is_connected:
            log.info("Actually connected now.")
        return is_connected

    def remove_client(self, uuid: str):
        self.credentials = [
            server for server in self.credentials if server["uuid"] != uuid
        ]
        self.save_credentials()
        self._disconnect_client(uuid=uuid)

    def connect_client(self, server, do_retries=True):
        if self.is_stopping:
            return False

        is_logged_in = False
        client = self.client_factory()
        state = client.authenticate({"Servers": [server]}, discover=False)
        server["connected"] = state["State"] == CONNECTION_STATE["SignedIn"]
        if server["connected"]:
            is_logged_in = self.setup_client(client, server)
            if is_logged_in:
                self.clients[server["uuid"]] = client
                if server.get("username"):
                    self.usernames[server["uuid"]] = server["username"]
            elif do_retries:
                # Jellyfin client sometimes "connects" halfway but doesn't actually work.
                # Retry three times to reduce odds of this happening.
                partial_reconnect_attempts = 3
                for i in range(partial_reconnect_attempts):
                    log.warning(
                        f"Partially connected. Retrying {i+1}/{partial_reconnect_attempts}."
                    )
                    self._disconnect_client(server=server)
                    time.sleep(1)
                    if self.connect_client(server, False):
                        is_logged_in = True
                        break

        return is_logged_in

    def _disconnect_client(self, uuid: Optional[str] = None, server=None):
        if uuid is None and server is not None:
            uuid = server["uuid"]

        if uuid not in self.clients:
            return

        if server is not None:
            server["connected"] = False

        client = self.clients[uuid]
        del self.clients[uuid]
        client.stop()

    def remove_all_clients(self):
        self.stop_all_clients()
        self.credentials = []
        self.save_credentials()

    def stop_all_clients(self):
        for key, client in list(self.clients.items()):
            del self.clients[key]
            client.stop()

    def check_all_clients(self):
        log.info("Performing client health check...")
        # list() because validate_client may mutate self.clients via the
        # synthesized WebSocketDisconnect path.
        for client in list(self.clients.values()):
            self.validate_client(client)
        # Retry credentials that aren't currently connected. Without this, a
        # server that fails the initial connect (e.g. shim started before LAN
        # was up) is never tried again until the user restarts the app —
        # the long-standing reliability hole behind issues #344 / #410.
        for server in self.credentials:
            if server["uuid"] not in self.clients and not self.is_stopping:
                log.info(
                    "Health check: retrying disconnected server %s",
                    server.get("address"),
                )
                self.connect_client(server, do_retries=False)

    def stop(self):
        if self.health_check:
            self.health_check.stop()
            self.health_check = None

        self.is_stopping = True
        for client in self.clients.values():
            client.stop()

    def get_username_from_client(self, client):
        # This is kind of convoluted. It may fail if a server
        # was added before we started saving usernames.
        for uuid, client2 in self.clients.items():
            if client2 is client:
                if uuid in self.usernames:
                    return self.usernames[uuid]
                for server in self.credentials:
                    if server["uuid"] == uuid:
                        return server.get("username", "Unknown")
                break

        return "Unknown"


clientManager = ClientManager()



import logging
import os

from .conf import settings
from .media import Media
from .player import playerManager
from .timeline import timelineManager

log = logging.getLogger("event_handler")
bindings = {}

NAVIGATION_DICT = {
    "Back": "back",
    "Select": "ok",
    "MoveUp": "up",
    "MoveDown": "down",
    "MoveRight": "right",
    "MoveLeft": "left",
    "GoHome": "home",
    "GoToSettings": "home",
}

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from jellyfin_apiclient_python import JellyfinClient as JellyfinClient_type


def bind(event_name: str):
    def decorator(func):
        bindings[event_name] = func
        return func

    return decorator


class EventHandler(object):
    mirror = None

    def handle_event(
        self,
        client: "JellyfinClient_type",
        event_name: str,
        arguments: dict,
    ):
        if event_name in bindings:
            log.debug("Handled Event {0}: {1}".format(event_name, arguments))
            bindings[event_name](self, client, event_name, arguments)
        else:
            log.debug("Unhandled Event {0}: {1}".format(event_name, arguments))

    @bind("Play")
    def play_media(self, client: "JellyfinClient_type", _event_name, arguments: dict):
        play_command = arguments.get("PlayCommand")
        if not playerManager.has_video():
            play_command = "PlayNow"

        if play_command == "PlayNow":
            seq = arguments.get("StartIndex")
            if seq is None:
                seq = 0
            media = Media(
                client,
                arguments.get("ItemIds"),
                seq=seq,
                user_id=arguments.get("ControllingUserId"),
                aid=arguments.get("AudioStreamIndex"),
                sid=arguments.get("SubtitleStreamIndex"),
                srcid=arguments.get("MediaSourceId"),
            )

            log.debug("EventHandler::playMedia %s" % media)
            offset = arguments.get("StartPositionTicks")
            if offset is not None:
                offset /= 10000000

            video = media.video
            if video:
                if settings.pre_media_cmd:
                    os.system(settings.pre_media_cmd)
                playerManager.play(video, offset, is_initial_play=True)
                timelineManager.send_timeline()
                if arguments.get("SyncPlayGroup") is not None:
                    playerManager.syncplay.join_group(arguments["SyncPlayGroup"])
                if settings.play_cmd:
                    os.system(settings.play_cmd)
        elif play_command == "PlayLast":
            playerManager.get_video().parent.insert_items(
                arguments.get("ItemIds"), append=True
            )
            playerManager.upd_player_hide()
        elif play_command == "PlayNext":
            playerManager.get_video().parent.insert_items(
                arguments.get("ItemIds"), append=False
            )
            playerManager.upd_player_hide()

    @bind("GeneralCommand")
    def general_command(
        self, client: "JellyfinClient_type", _event_name, arguments: dict
    ):
        command = arguments.get("Name")
        if command == "SetVolume":
            # There is currently a bug that causes this to be spammed, so we
            # only update it if the value actually changed.
            if playerManager.get_volume(True) != int(arguments["Arguments"]["Volume"]):
                playerManager.set_volume(int(arguments["Arguments"]["Volume"]))
        elif command == "SetAudioStreamIndex":
            playerManager.set_streams(int(arguments["Arguments"]["Index"]), None)
        elif command == "SetSubtitleStreamIndex":
            playerManager.set_streams(None, int(arguments["Arguments"]["Index"]))
        elif command == "DisplayContent":
            # If you have an idle command set, this will delay it.
            timelineManager.delay_idle()
            if self.mirror:
                self.mirror.display_content(client, arguments)
        elif command in (
            "Back",
            "Select",
            "MoveUp",
            "MoveDown",
            "MoveRight",
            "MoveLeft",
            "GoHome",
            "GoToSettings",
        ):
            playerManager.menu_action(NAVIGATION_DICT[command])
        elif command in ("Mute", "Unmute"):
            playerManager.set_mute(command == "Mute")
        elif command == "TakeScreenshot":
            playerManager.screenshot()
        elif command == "ToggleFullscreen" or command is None:
            # Currently when you hit the fullscreen button, no command is specified...
            playerManager.toggle_fullscreen()

    @bind("Playstate")
    def play_state(self, _client: "JellyfinClient_type", _event_name, arguments: dict):
        command = arguments.get("Command")
        if command == "PlayPause":
            playerManager.toggle_pause()
            timelineManager.send_timeline()
        elif command == "Pause":
            playerManager.pause_if_playing()
            timelineManager.send_timeline()
        elif command == "Unpause":
            playerManager.play_if_paused()
            timelineManager.send_timeline()
        elif command == "PreviousTrack":
            playerManager.play_prev()
        elif command == "NextTrack":
            playerManager.play_next()
        elif command == "Stop":
            playerManager.stop()
        elif command == "Seek":
            playerManager.seek(
                arguments.get("SeekPositionTicks") / 10000000, absolute=True
            )

    @bind("PlayPause")
    def pause_play(self, _client: "JellyfinClient_type", _event_name, _arguments: dict):
        playerManager.toggle_pause()
        timelineManager.send_timeline()

    @bind("SyncPlayGroupUpdate")
    def sync_play_group_update(
        self, client: "JellyfinClient_type", _event_name, arguments: dict
    ):
        playerManager.syncplay.client = client
        playerManager.syncplay.process_group_update(arguments)

    @bind("SyncPlayCommand")
    def sync_play_command(
        self, client: "JellyfinClient_type", _event_name, arguments: dict
    ):
        playerManager.syncplay.client = client
        playerManager.syncplay.process_command(arguments)


eventHandler = EventHandler()



APP_NAME = "jellyfin-mpv-shim"
USER_APP_NAME = "Jellyfin MPV Shim"
CLIENT_VERSION = "2.10.0"
USER_AGENT = "Jellyfin-MPV-Shim/%s" % CLIENT_VERSION
CAPABILITIES = {
    "PlayableMediaTypes": ["Video"],
    "SupportsMediaControl": True,
    "SupportsPersistentIdentifier": True,
    "SupportedCommands": [
        "MoveUp",
        "MoveDown",
        "MoveLeft",
        "MoveRight",
        "Select",
        "Back",
        "ToggleFullscreen",
        "GoHome",
        "GoToSettings",
        "TakeScreenshot",
        "VolumeUp",
        "VolumeDown",
        "ToggleMute",
        "SetAudioStreamIndex",
        "SetSubtitleStreamIndex",
        "Mute",
        "Unmute",
        "SetVolume",
        "DisplayContent",
        "Play",
        "Playstate",
        "PlayNext",
        "PlayMediaSource",
    ],
}



"""Single source of truth for command-line arguments.

Other modules read parsed values via get_args(); they should never scan sys.argv
directly. Adding a new flag means editing this file and nothing else.
"""

import argparse

from .constants import APP_NAME, CLIENT_VERSION

_args = None


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=APP_NAME,
        description="Jellyfin MPV Shim - cast media from Jellyfin to MPV.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + CLIENT_VERSION,
    )
    parser.add_argument(
        "--config",
        metavar="DIR",
        help="use a custom configuration directory",
    )
    parser.add_argument(
        "--gui",
        dest="enable_gui",
        action=argparse.BooleanOptionalAction,
        default=None,
        help="enable or disable the GUI (overrides config)",
    )
    parser.add_argument(
        "--mpv-loglevel",
        dest="mpv_loglevel",
        choices=("fatal", "error", "warn", "info", "debug"),
        default=None,
        help="override mpv_log_level for this run",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="surface debug-level application log messages (does not affect mpv verbosity)",
    )
    parser.add_argument(
        "--server",
        metavar="URL",
        help="server URL for non-interactive credential add (requires --username)",
    )
    parser.add_argument(
        "--username",
        metavar="NAME",
        help="username for --server",
    )
    parser.add_argument(
        "--password",
        metavar="PASS",
        default="",
        help="password for --server (NOTE: visible to other processes via ps)",
    )
    parser.add_argument(
        "command",
        nargs="*",
        choices=("add", "clear"),
        help="add: prompt to add a server; clear: remove all stored credentials",
    )
    return parser


def get_args() -> argparse.Namespace:
    """Parse argv on first call, cache the result thereafter."""
    global _args
    if _args is None:
        _args = _build_parser().parse_args()
    return _args



from .conf import settings
from . import conffile
from .utils import get_resource
from .constants import APP_NAME
from .i18n import _
import logging
import os.path
import shutil
import json

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .player import PlayerManager as PlayerManager_type
    from .menu import OSDMenu as OSDMenu_type

profile_name_translation = {
    "Generic (FSRCNNX)": _("Generic (FSRCNNX)"),
    "Generic High (FSRCNNX x16)": _("Generic High (FSRCNNX x16)"),
    "Anime4K x4 Faithful (For SD)": _("Anime4K x4 Faithful (For SD)"),
    "Anime4K x4 Perceptual (For SD)": _("Anime4K x4 Perceptual (For SD)"),
    "Anime4K x4 Perceptual + Deblur (For SD)": _(
        "Anime4K x4 Perceptual + Deblur (For SD)"
    ),
    "Anime4K x2 Faithful (For HD)": _("Anime4K x2 Faithful (For HD)"),
    "Anime4K x2 Perceptual (For HD)": _("Anime4K x2 Perceptual (For HD)"),
    "Anime4K x2 Perceptual + Deblur (For HD)": _(
        "Anime4K x2 Perceptual + Deblur (For HD)"
    ),
}

log = logging.getLogger("video_profile")


class MPVSettingError(Exception):
    """Raised when MPV does not support a required setting."""

    pass


class VideoProfileManager:
    def __init__(
        self, menu: "OSDMenu_type", player_manager: "PlayerManager_type", player
    ):
        self.menu = menu
        self.playerManager = player_manager
        self.used_settings = set()
        self.current_profile = None
        self.player = player
        self.profile_subtypes = []

        shader_pack_builtin = get_resource("default_shader_pack")

        # Load shader pack
        self.shader_pack = shader_pack_builtin
        if settings.shader_pack_custom:
            self.shader_pack = conffile.get(APP_NAME, "shader_pack")
            if not os.path.exists(self.shader_pack):
                shutil.copytree(shader_pack_builtin, self.shader_pack)

        pack_name = "pack-next.json"
        if not os.path.exists(os.path.join(self.shader_pack, pack_name)):
            pack_name = "pack.json"

            if not os.path.exists(os.path.join(self.shader_pack, pack_name)):
                raise FileNotFoundError("Could not find default shader pack.")

        with open(os.path.join(self.shader_pack, pack_name)) as fh:
            pack = json.load(fh)
            self.default_groups = pack.get("default-setting-groups") or []
            self.profiles = pack.get("profiles") or {}
            self.groups = pack.get("setting-groups") or {}
            self.revert_ignore = set(pack.get("setting-revert-ignore") or [])

            self.profile_subtypes = set()
            for profile in self.profiles.values():
                for subtype in profile.get("subtype", []):
                    self.profile_subtypes.add(subtype)

        self.defaults = {}
        for group in self.groups.values():
            setting_group = group.get("settings")
            if setting_group is None:
                continue

            for key, value in setting_group:
                if key in self.defaults or key in self.revert_ignore:
                    continue
                try:
                    self.defaults[key] = getattr(self.player, key)
                except Exception:
                    log.warning(
                        "Your MPV does not support setting {0} used in shader pack.".format(
                            key
                        ),
                        exc_info=True,
                    )

        if settings.shader_pack_profile is not None:
            self.load_profile(settings.shader_pack_profile, reset=False)

    def process_setting_group(
        self, group_name: str, settings_to_apply: list, shaders_to_apply: list
    ):
        group = self.groups[group_name]
        for key, value in group.get("settings", []):
            if key not in self.defaults:
                if key not in self.revert_ignore:
                    raise MPVSettingError(
                        "Cannot use setting group {0} due to MPV not supporting {1}".format(
                            group_name, key
                        )
                    )
            else:
                self.used_settings.add(key)
            settings_to_apply.append((key, value))
        for shader in group.get("shaders", []):
            shaders_to_apply.append(os.path.join(self.shader_pack, "shaders", shader))

    def load_profile(self, profile_name: str, reset: bool = True):
        if reset:
            self.unload_profile()
        log.info("Loading shader profile {0}.".format(profile_name))
        if profile_name not in self.profiles:
            log.error("Shader profile {0} does not exist.".format(profile_name))
            return False

        profile = self.profiles[profile_name]
        settings_to_apply = []
        shaders_to_apply = []
        try:
            # Read Settings & Shaders
            for group in self.default_groups:
                self.process_setting_group(group, settings_to_apply, shaders_to_apply)
            for group in profile.get("setting-groups", []):
                self.process_setting_group(group, settings_to_apply, shaders_to_apply)
            for shader in profile.get("shaders", []):
                shaders_to_apply.append(
                    os.path.join(self.shader_pack, "shaders", shader)
                )

            # Apply Settings
            already_set = set()
            for key, value in settings_to_apply:
                if (key, value) in already_set:
                    continue
                log.info("Set MPV setting {0} to {1}".format(key, value))
                setattr(self.player, key, value)
                already_set.add((key, value))

            # Apply Shaders
            log.info("Set shaders: {0}".format(shaders_to_apply))
            self.player.glsl_shaders = shaders_to_apply
            self.current_profile = profile_name
            return True
        except MPVSettingError:
            log.error("Could not apply shader profile.", exc_info=True)
            return False

    def unload_profile(self):
        log.info("Unloading shader profile.")
        self.player.glsl_shaders = []
        for setting in self.used_settings:
            value = self.defaults[setting]
            try:
                setattr(self.player, setting, value)
            except Exception:
                log.warning(
                    "Default setting {0} value {1} is invalid.".format(setting, value)
                )
        self.current_profile = None

    def menu_handle(self):
        profile_name = self.menu.menu_list[self.menu.menu_selection][2]
        settings_were_successful = True
        if profile_name is None:
            self.unload_profile()
        else:
            settings_were_successful = self.load_profile(profile_name)
        if settings.shader_pack_remember and settings_were_successful:
            settings.shader_pack_profile = profile_name
            settings.save()

        # Need to re-render menu.
        self.menu.menu_action("back")
        self.menu_action()

    def menu_action(self):
        selected = 0
        profile_option_list = [(_("None (Disabled)"), self.menu_handle, None)]
        for i, (profile_name, profile) in enumerate(self.profiles.items()):
            if (
                profile.get("subtype", None) is not None
                and not settings.shader_pack_subtype in profile["subtype"]
            ):
                continue

            name = profile["displayname"]
            if name in profile_name_translation:
                name = profile_name_translation[name]
            profile_option_list.append((name, self.menu_handle, profile_name))
            if profile_name == self.current_profile:
                selected = i + 1
        self.menu.put_menu(_("Select Shader Profile"), profile_option_list, selected)
